%  This code can be used to perform power flow assuming balancing LV
%  microgrids, where each DG unit is equipped with P(f) and Q(V) droop
%  characteristics. More details regarding the examined network are
%  presented in Ref. [1] below.
%
% When using this code please cite one of the following:
% 
% [1] E. O. Kontis, G. C. Kryonidis, A. I. Nousdilis, K.-N. D. Malamaki, and G. K. Papagiannis, 
% "Power flow of islanded AC microgrids," IEEE Milano PowerTech 2019, Milano, Italy, June 23 - 27, 2019.
% 
% [2] G. C. Kryonidis, E. O. Kontis, A. I. Chrysochos, K. O. Oureilidis, C. S. Demoulias, 
% and G. K. Papagiannis, % "Power flow of islanded AC microgrids: Revisited," IEEE Transactions on 
% Smart Grid, vol. 9, no. 4, pp. 3903 - 3905, July 2018.
%
% [3] G. C. Kryonidis, E. O. Kontis, A. I. Chrysochos, C. S. Demoulias, D. Bozalakov, B. Meersman, 
% T. L. Vandoorn, and L. Vandevelde, "A simulation tool for extended distribution grids with 
% controlled distributed generation," Proceedings of the IEEE Eindhoven PowerTech 2015, Eindhoven, 
% Netherlands, June 29 - July 2, 2015.
% 
% Contact information: 
% Prof. Grigoris K. Papagiannis 
% School of Electrical and Computer Engineering
% Aristotle University of Thessaloniki
% e-mail: grigoris@eng.auth.gr

%% Simulation parameters

load('Network Data.mat')                      % Network electrical characteristics
Pmpp = [40000,30000,40000,50000,50000,30000]; % Maximum point power (MPP) of each DG unit in W
intgrl = 2;                                   % Integral term k
sb_location = 1;                              % Location of the virtual slack bus
fmin = 49;                                    % Minimum permissible frequency in Hz
fmax = 51;                                    % Maximum permissible frequency in Hz
Vmin = 0.9*0.4;                               % Minimum permissible voltage of the virtual slack bus
Vmax = 1.1*0.4;                               % Maximum permissible voltage of the virtual slack bus
Qmin = -tan(acos(0.85))*Pmpp;                 % Minimum reactive power of each DG unit in Var
Qmax = tan(acos(0.85))*Pmpp;                  % Maximum reactive power of each DG unit in Var
tol.active = 0.1;                             % Tolerance of the active power loop in W
tol.reactive = 0.1;                           % Tolerance of the reactive power loop in Var
tol.volt =0.00001;                            % Tolerance of the integral term loop in V

%% Initial Values

Ploss = 0;                                    % Network losses (active and reactive power) in W
fmicrogrid = 51;                              % Microgrid frequency in Hz
Vgrid = Vmax;                                 % Voltage of the slack bus in kV
Qgen = Qmin;                                  % Reactive power of each DG unit in Var
Pgen = zeros(1,6);                            % Active power of each DG unit in W
Vup = Vmax;                                   % Auxiliary variable for the binary search algorithm
Vlow = Vmin;                                  % Auxiliary variable for the binary search algorithm
Vmid = Vmax;                                  % Auxiliary variable for the binary search algorithm

%% Initial code

coef.p = Pmpp/min(Pmpp);                      % Auxiliary variable
tic                                           % Starting point of the execution time
data_write(fmicrogrid,Vgrid,sb_location);     % Write the .dss file for OpenDSS
Line(18,2) = sb_location;                     % Assign a virtual line between the virtual slack bus and the MG node

DSSObj = actxserver('OpenDSSEngine.DSS');                    %\
Start = DSSObj.Start(0);                                     % \
DSSCircuit = DSSObj.ActiveCircuit;                           %  |-> Initialization of the OpenDSS
DSSText = DSSObj.Text;                                       % /
DSSText.Command = ['compile ',pwd,'\Examined_Network.dss'];  %/

[rows,~] = size(Gen);                                              %\
for i = 1 : rows                                                   % \
    DSSCircuit.SetActiveElement(['generator.',Genname{i},'a']);    %  \
    DSSCircuit.ActiveElement.Properties('kW').Val=num2str(0);      %   \
    DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(0);    %    \
                                                                   %     \
    DSSCircuit.SetActiveElement(['generator.',Genname{i},'b']);    %      \
    DSSCircuit.ActiveElement.Properties('kW').Val=num2str(0);      %       |-> Set the active and reactive power output of each DG unit
    DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(0);    %      /
                                                                   %     /
    DSSCircuit.SetActiveElement(['generator.',Genname{i},'c']);    %    /
    DSSCircuit.ActiveElement.Properties('kW').Val=num2str(0);      %   /
    DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(0);    %  /
end                                                                % /
                                                                   %/

% First Load Flow Solution
DSSSolution = DSSCircuit.Solution;
DSSSolution.Solve;

% Results extraction and calculation
if DSSSolution.Converged == 1                                                                                       % Check whether the power flon in OpenDSS has been converged
    Ploss = DSSCircuit.Losses;                                                                                      % Calculate network losses
    Voltage = voltage_extraction(Busname,DSSCircuit.AllNodename,DSSCircuit.AllBusVolts);                            % Calculate network voltages
    DSSCircuit.SetActiveElement('line.phantom_line');                                                               %\
    curline1 = DSSCircuit.ActiveElement.Currents;                                                                   % \
    cur.pha = curline1(1)+1i*curline1(2);                                                                           %  \
    cur.phb = curline1(3)+1i*curline1(4);                                                                           %   \
    cur.phc = curline1(5)+1i*curline1(6);                                                                           %    |-> Calculate the active and reactive power flowing through the slack bus
    cur.phn = curline1(7)+1i*curline1(8);                                                                           %   /
    Sfeeder = Voltage.VAN_vec(1)*conj(cur.pha)+Voltage.VBN_vec(1)*conj(cur.phb)+Voltage.VCN_vec(1)*conj(cur.phc);   %  /
    Pfeeder = real(Sfeeder);                                                                                        % /
    Qfeeder = imag(Sfeeder);                                                                                        %/
else
    display('OpenDSS cannot converge.');      
end


%% Microgrid Simulation (MAIN CODE)
% Reactive power loop
while abs(Qfeeder) > tol.reactive || abs(Pfeeder) > tol.active

    p_f_droop_flag = 1;                                                 % Auxiliary variable for the termination of active power loop
    q_v_droop_flag = 1;                                                 % Auxiliary variable for the termination of reactive power loop
    int_loop_flag = 1;                                                  % Auxiliary variable for the termination of integral term loop
    data_write(fmicrogrid,Vgrid,sb_location);                           % Write the .dss file for OpenDSS
    DSSText.Command = ['compile ',pwd,'\Examined_Network.dss'];
 
    % Active power loop
    while p_f_droop_flag
        [rows,~] = size(Gen);                                                              %\
        for i = 1 : rows                                                                   % \
                DSSCircuit.SetActiveElement(['generator.',Genname{i},'a']);                %  \
                DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);    %   \
                DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);  %    \
                                                                                           %     \ 
                DSSCircuit.SetActiveElement(['generator.',Genname{i},'b']);                %      \
                DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);    %       |-> Set the active and reactive power output of each DG unit
                DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);  %      /
                                                                                           %     /
                DSSCircuit.SetActiveElement(['generator.',Genname{i},'c']);                %    /
                DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);    %   /
                DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);  %  /
        end                                                                                % /
                                                                                           %/
 
        % Load Flow Solution
        DSSSolution = DSSCircuit.Solution;
        DSSSolution.Solve;

        % Results extraction and calculation
        if DSSSolution.Converged == 1                                                                                        % Check whether the power flon in OpenDSS has been converged
            Ploss = DSSCircuit.Losses;                                                                                       % Calculate network losses
            Voltage = voltage_extraction(Busname,DSSCircuit.AllNodename,DSSCircuit.AllBusVolts);                             % Calculate network voltages
            DSSCircuit.SetActiveElement('line.phantom_line');                                                                %\
            curline1 = DSSCircuit.ActiveElement.Currents;                                                                    % \
            cur.pha = curline1(1)+1i*curline1(2);                                                                            %  \
            cur.phb = curline1(3)+1i*curline1(4);                                                                            %   \
            cur.phc = curline1(5)+1i*curline1(6);                                                                            %    |-> Calculate the active and reactive power flowing through the slack bus
            cur.phn = curline1(7)+1i*curline1(8);                                                                            %   /
            Sfeeder = Voltage.VAN_vec(1)*conj(cur.pha)+Voltage.VBN_vec(1)*conj(cur.phb)+Voltage.VCN_vec(1)*conj(cur.phc);    %  /
            Pfeeder = real(Sfeeder);                                                                                         % /  
            Qfeeder = imag(Sfeeder);                                                                                         %/
        else
            display('OpenDSS cannot converge.');      
        end
        
        if abs(Pfeeder) > tol.active
            % Implementation of the P-f droop control
            Pprov = (Pfeeder)/sum(coef.p);                                %\
            Pgen = Pprov*coef.p+Pgen;                                     % |-> Calculate the new Microgrid frequency and active power output of each DG unit
            fmicrogrid = fmax-Pgen(1)*(fmax-fmin)/Pmpp(1);                %/
            data_write(fmicrogrid,Vgrid,sb_location);                     % Write the .dss file for OpenDSS
            DSSText.Command = ['compile ',pwd,'\Examined_Network.dss'];
        else
            p_f_droop_flag=0;
        end
        
        
    end

    % Initial conditions of the reactive power of each DG unit 
    Qgen(1)=Qmin(1); Qgen(2)=Qmin(2); Qgen(3)=Qmin(3);   
    Qgen(4)=Qmin(4); Qgen(5)=Qmin(5);   

    % Loop with the integral term
    while int_loop_flag
                                                                                                                %\
        Qgen_temp(1) = Qmin(1) + (sqrt(3)*abs(Voltage.VAN_vec(2))/1000-Vmax)*(Qmax(1)-Qmin(1))/(Vmin-Vmax);     % \
        Qgen_temp(2) = Qmin(2) + (sqrt(3)*abs(Voltage.VAN_vec(5))/1000-Vmax)*(Qmax(2)-Qmin(2))/(Vmin-Vmax);     %  \
        Qgen_temp(3) = Qmin(3) + (sqrt(3)*abs(Voltage.VAN_vec(7))/1000-Vmax)*(Qmax(3)-Qmin(3))/(Vmin-Vmax);     %   |-> Calculate the reactive power output of each DG unit based on the Q(V) droop and the voltage at the PCC
        Qgen_temp(4) = Qmin(4) + (sqrt(3)*abs(Voltage.VAN_vec(10))/1000-Vmax)*(Qmax(4)-Qmin(4))/(Vmin-Vmax);    %  /
        Qgen_temp(5) = Qmin(5) + (sqrt(3)*abs(Voltage.VAN_vec(12))/1000-Vmax)*(Qmax(5)-Qmin(5))/(Vmin-Vmax);    % /
        Qgen_temp(6) = Qmin(6) + (sqrt(3)*abs(Voltage.VAN_vec(15))/1000-Vmax)*(Qmax(6)-Qmin(6))/(Vmin-Vmax);    %/
        
                                                                 %\
        Qgen(1) = Qgen(1) + (Qgen_temp(1)-Qgen(1))/intgrl;       % \         
        Qgen(2) = Qgen(2) + (Qgen_temp(2)-Qgen(2))/intgrl;       %  \
        Qgen(3) = Qgen(3) + (Qgen_temp(3)-Qgen(3))/intgrl;       %   |-> Calculate the reactive power that will be applied to each DG unit with the use of the integral term
        Qgen(4) = Qgen(4) + (Qgen_temp(4)-Qgen(4))/intgrl;       %  /
        Qgen(5) = Qgen(5) + (Qgen_temp(5)-Qgen(5))/intgrl;       % /
        Qgen(6) = Qgen(6) + (Qgen_temp(6)-Qgen(6))/intgrl;       %/
        
        DSSText.Command = ['compile ',pwd,'\Examined_Network.dss'];
        
        [rows,~]=size(Gen);                                                              %\
        for i=1:rows                                                                     % \
            DSSCircuit.SetActiveElement(['generator.',Genname{i},'a']);                  %  \
            DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);      %   \
            DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);    %    \
                                                                                         %     \
            DSSCircuit.SetActiveElement(['generator.',Genname{i},'b']);                  %      \
            DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);      %       |->% Set the active and reactive power output of each DG unit
            DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);    %      /
                                                                                         %     /
            DSSCircuit.SetActiveElement(['generator.',Genname{i},'c']);                  %    /
            DSSCircuit.ActiveElement.Properties('kW').Val=num2str(Pgen(i)/3000,10);      %   /
            DSSCircuit.ActiveElement.Properties('kvar').Val=num2str(Qgen(i)/3000,10);    %  /
        end                                                                              % /
                                                                                         %/
        
        % Load Flow Solution
        DSSSolution = DSSCircuit.Solution;
        DSSSolution.Solve;

        % Results extraction
        if DSSSolution.Converged == 1                                                                                                     % Check whether the power flon in OpenDSS has been converged
            Ploss = DSSCircuit.Losses;                                                                                                    % Calculate network losses
            Voltage_temp = voltage_extraction(Busname,DSSCircuit.AllNodename,DSSCircuit.AllBusVolts);                                     % Calculate network voltages
            DSSCircuit.SetActiveElement('line.phantom_line');                                                                             %\
            curline1 = DSSCircuit.ActiveElement.Currents;                                                                                 % \
            cur.pha = curline1(1)+1i*curline1(2);                                                                                         %  \
            cur.phb = curline1(3)+1i*curline1(4);                                                                                         %   \
            cur.phc = curline1(5)+1i*curline1(6);                                                                                         %    |-> Calculate the active and reactive power flowing through the slack bus
            cur.phn = curline1(7)+1i*curline1(8);                                                                                         %   /
            Sfeeder = Voltage_temp.VAN_vec(1)*conj(cur.pha)+Voltage_temp.VBN_vec(1)*conj(cur.phb)+Voltage_temp.VCN_vec(1)*conj(cur.phc);  %  /
            Pfeeder = real(Sfeeder);                                                                                                      % /
            Qfeeder = imag(Sfeeder);                                                                                                      %/
        else
            display('OpenDSS cannot converge.');      
        end
        
        
        temp = 0;                                                            %\
        for i = 1 : length(Voltage.VAN_vec)                                  % \
            if abs(Voltage.VAN_vec(i)-Voltage_temp.VAN_vec(i)) < tol.volt    %  \
                temp = temp + 1;                                             %   \
            end                                                              %    \
        end                                                                  %     |-> Termination check of the intergral term loop
                                                                             %    /
        if temp == length(Voltage.VAN_vec)                                   %   /
           int_loop_flag = 0;                                                %  /
        end                                                                  % /
        Voltage = Voltage_temp;                                              %/
    end
    
    if Qfeeder > 0                                                           %\
        Vgrid = Vmid - (Vmid-Vlow)/2;                                        % \
        Vup = Vmid;                                                          %  \
        Vmid = Vgrid;                                                        %   \
                                                                             %    |-> Determine the new voltage of the slack bus using the binary search algorithm         
    else                                                                     %   /
        Vgrid = Vmid + (Vup-Vmid)/2;                                         %  /
        Vlow = Vmid;                                                         % /
        Vmid = Vgrid;                                                        %/
    end
    
    
end

DSSObj.ClearAll;                                                             % Clear DSS object
time = toc;                                                                  % Ending point of the execution time

% Results save
save([pwd,filesep,'Results',filesep,'Results - slack bus at ',Busname{sb_location},' - integral = ',num2str(intgrl),'.mat'],'Sfeeder','Pgen','Qgen','Ploss','Vgrid','Voltage','Busname','Genname','fmicrogrid','time')
